% LICENSE AGREEMENT
% - - - - - - - - -
% Copyright (c) 2019, Stilianos Louca
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are
% met:
%
%    * Redistributions of source code must retain the above copyright
%      notice, this list of conditions and the following disclaimer.
%    * Redistributions in binary form must reproduce the above copyright
%      notice, this list of conditions and the following disclaimer in
%      the documentation and/or other materials provided with the distribution
%    * Neither the name of the original author (Stilianos Louca), nor the names
%      of any other contributors may be used to endorse or promote products derived
%      from this software without specific prior written permission.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
% ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
% LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%
% - - - - - - - - -
% Code name and version: SMF_example_ASOS 2019.01.06
% Tested on MATLAB R2015b.
% - - - - - - - - -
%
% DESCRIPTION
% - - - - - -
% This MATLAB code demonstrates Spatial Metabolic Flux analysis for the following simple steady-state 1-dimensional scenario:
%	Sulfide (H2S) diffusing from the bottom boundary of a water column is completely oxidized 
%	using oxygen (O2) diffusing from the top boundary, within a single hotspot, at a stoichiometric ratio H2S:O2 = 1:2.
%	No other reaction is assumed to occur.
%	Metabolite concentrations (H2S and O2) at the top and bottom boundaries are assumed to be fixed.
%	The resulting steady-state metabolite concentrations profiles across depth, the vertical metabolite fluxes and the reaction rate are all predicted via SMF.
%
% The code performs the following main tasks:
%   1. Specifies the model (including metabolites, reactions, stoichiometric matrix and diffusivity) for the SMF analysis
%   2. Runs the SMF analysis, predicting the hotspot location and reaction rate. 
%   3. Plots the predicted H2S and O2 depth profiles and prints out the predicted reaction rate.
% Task 2 is performed by the accompanied SMF workflow in file "fitSMFmodel.m".
% Task 1 is performed in this source file, while task 3 is performed using auxiliary plotting routines provided in separate source files ("plotProfiles.m" and "num2RowsColumns.m").
% In general, you can use your own code to perform tasks 1 and 3.
%
% Time is in days, depth in meters (m), metabolite concentrations in mol/litre (M), diffusivity in m^2/day, reaction rates in kmol/m^2/day.
%
% Tested on MATLAB R2015b.
% To run this code, simply run 'main' in the MATLAB console.
% 
% Stilianos Louca
% February 12, 2019
%
function main()

	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	% TECHNICAL PREPARATIONS & OPTIONS
	
	% figure out uniquely numbered output dir
	outputDir = getNonExistingSimilarDir('output', 'example_run_');
	makedir(outputDir);
			
	%output precision
	format long;
		
	fprintf('Initializing..\nNote: All outputs and plots will be saved in the directory "%s"\n', outputDir);
	
	PLOT_WIDTH	= 3;	% depth-profile plot width, in inches
	PLOT_HEIGHT	= 3;	% depth-profile plot height, in inches

	
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	% SPECIFY MODEL FOR SMF ANALYSIS
	
	% define considered metabolites & reactions
	% Products not limiting to any reaction (such as SO4-2 or H+) can be ignored
	% Element and electron balancing by each reaction is not required for SMF analysis
	reaction_names 	 = {'ASOS'};
	metabolite_names = {'H2S' 'O2'};
	
	% define reaction stoichiometry
	% This 2D matrix must be consistent with the metabolites & reactions defined above
	% The entry in row m and column r is the stoichiometric coefficient of metabolite m in reaction r
	stoichiometry =	[ 	-1;		% H2S
						-2];	% O2

	% check if dimensions are correct
	NR = numel(reaction_names);
	NM = numel(metabolite_names);
	if((size(stoichiometry,1)~=NM) || (size(stoichiometry,2)~=NR))
		fprintf('ERROR: Stoichiometric matrix must have %d rows and %d columns, but instead has %d rows and %d columns\n', NM, NR, size(stoichiometry,1), size(stoichiometry,2));
		return;
	end
	
	% specify boundary concentrations (in mol/L) at top & bottom of the water column
	% at each boundary (top & bottom), boundary concentrations should be 1D vectors of size NM
	concentrations_top = [	0;			% H2S
							50e-6]; 	% O2
	concentrations_bottom = [	10e-6;	% H2S
								0]; 	% O2
							
	% specify the modeled number of hotspots
	NH = 1;
	
	% specify the depth-grid to be considered, including the top and bottom boundary
	% a finer grid increases accuracy, but also increases computation time
	SMF_DEPTHS = linspace(0, 1000, 200);
	NZ = numel(SMF_DEPTHS);

	% Specify the eddy diffusivity for all depths
	% The diffusivity should be defined as a function of depth
	% Feel free to define an alternative diffusivity profile, but make sure it's strictly positive everywhere
	% In this example, the diffusivity is assumed to be the same for all metabolites, and to increase linearly with depth
	diffusivity = @(z) 1+10*z./SMF_DEPTHS(end);
	
	% plot diffusivity profile
	fprintf('Plotting assumed diffusivity profile..\n');
	plotProfiles(	{diffusivity},...
					SMF_DEPTHS, ...
					{}, ...
					NaN, ...
					[SMF_DEPTHS(1),SMF_DEPTHS(end)], ...
					{'eddy diffusivity'}, ...
					'diffusivity $(m^2/day)$', ...
					strcat(outputDir,'/diffusivity.pdf'),...
					'Model specs', ...	% supertitle
					0,... 				% min x value
					NaN, ... 			% refValue
					false, ...			% logarithmic
					false, ...			% show_points
					PLOT_WIDTH, ...
					PLOT_HEIGHT, ...
					false, ... 	% showDepthIntegrated
					false, ...	% showExtrema
					true, ... 	% showR2P2
					NaN, ...	% data_colors
					strcat(outputDir,'/plot_data/diffusivity.tsv'));



	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	% Run SMF analysis
	
	fprintf('Running SMF between depths %g-%g m..\n',SMF_DEPTHS(1),SMF_DEPTHS(end));
		
	[hotspots,...
	reaction_rates,...
	~,...				% ghost_source_rates. Irrelevant in this example.
	~,...				% ghost_sink_rates. Irrelevant in this example.
	concentrations,...
	concentrations_BC_part,...
	~,...				$ concentrations_added_production_part. Irrelevant in this example.
	concentrations_reaction_part,...
	metabolite_net_source_rates,...
	metabolite_boundary_influxes,...
	stress,...
	~,... 				% R2s_grid_data. Irrelevant in this example.
	~,... 				% R2s_scatter_data. Irrelevant in this example.
	hotspots_all_trials,...
	reaction_rates_all_trials,...
	stress_all_trials,...
	Niterations_all_trials,...
	Nevaluations_all_trials,...
	~,... 				% R2s_grid_data_all_trials. Irrelevant in this example.
	~,... 				% R2s_scatter_data_all_trials. Irrelevant in this example.
	~,...				% spurious_hotspots. Irrelevant in this example.
	~,...				% neglibible_hotspots. Irrelevant in this example.
	exit_flag,...
	Niterations,...
	Nfailed_trials,...
	errormsg] = fitSMFmodel(SMF_DEPTHS,...				% gridZ,...
							NM,... 						% number of metabolites
							NR,...						% number of reactions
							NH,...						% number of hotspots
							metabolite_names,...
							reaction_names,...
							zeros(NM,1),...				% BCtypesA. All boundary conditions are Dirichlet type (fixed concentration).
							zeros(NM,1),...				% BCtypesB. All boundary conditions are Dirichlet type (fixed concentration).
							concentrations_top,...		% BCvaluesA
							concentrations_bottom,...	% BCvaluesB
							{diffusivity, diffusivity},...	% diffusivities. In this example, the same iffusivity is used for all metabolites.
							{},...						% advections. Advection is assumed to be zero for all metabolites in this example.
							@(z) 1,...					% dilution_area (as function of depth). Lateral (cross-sectional) basin area is assumed to be constant in this example, i.e. assume no geometric dilution effects.
							NaN,...						% grid_temperatures. Irrelevant in this example.
							true,...					% use_ficks_law
							stoichiometry,...			% reaction_stoichiometry
							NaN,...						% reaction_DeltaG0. Irrelevant in this example.
							{},...						% concentration_grid_data
							{},...						% concentration_scatter_data
							[],...						% added_production_rates,...
							[],...						% ghost_sources (optional)
							[],...						% ghost_sinks (optional)
							1,...						% metabolite_weights
							1e-6,...					% positivity_scales
							1e-6,...					% negativity_scales
							'substrate_geomean',...		% penalize_inefficiencies_by
							Inf,...						% negative_concentration_penalty
							0,...						% negative_flux_penalty
							false,...					% include_grid_data
							false,...					% include_scatter_data
							[],...						% start_hotspots (optional)
							[],...						% start_reaction_rates (optional)
							false,...					% always_use_start_hotspots
							false,...					% always_use_start_reaction_rates
							[],...						% fixed_hotspots (optional)
							[],...						% fixed_reaction_rates (optional)
							[-Inf +Inf],...				% allowed_hotspot_range (optional)
							'fmincon',...				% optim_method
							500,...						% optim_max_iter
							50000,...					% optim_max_evals
							1e-5,...					% objective_tolerance
							1e-8,...					% parameter_tolerance
							1e-12,...					% concentration_epsilon
							1000,...					% max_random_rate_factor
							50,...						% Ntrials
							1,...						% Nrounds
							10,...						% Nstart_attempts
							true,...					% parallel
							false,...					% omit_spurious_hotspots
							-1,...						% negligible_rel_flux_threshold
							false,...					% R2s_for_all_trials
							true,...					% check_nans_in_input
							true,...					% verbose
							'  ');						% verbose_prefix

	if(~strcmp(errormsg,''))
		% SMF failed
		fprintf('  ERROR:  SMF failed: %s\n',errormsg);
	else
		% SMF succeeded, so print summary and plot predictions.
		fprintf('  Finished SMF:\n    stress: %g\n    Nfailed_trials: %d\n    Niterations: %d\n',stress,Nfailed_trials,Niterations);
		fprintf('    Predicted hotspot depth: %g m\n',hotspots(1));
		fprintf('    Predicted %s reaction rate: %g mol/m^2/day\n',reaction_names{1},reaction_rates(1));
		fprintf('    Predicted net %s consumption rate: %g mol/m^2/day\n',metabolite_names{1},-metabolite_net_source_rates(1));
	
		fprintf('Plotting SMF-predicted metabolite concentrations..\n');
		plotProfiles(	concentrations',...
						SMF_DEPTHS, ...
						{}, ...
						NaN, ...
						[SMF_DEPTHS(1),SMF_DEPTHS(end)], ...
						metabolite_names, ...
						'concentration (mol/L)', ...
						strcat(outputDir,'/predicted_metabolite_concentrations.pdf'),...
						sprintf('SMF-predicted metabolite concentrations (stress=%g)',stress), ...
						NaN,... 		% min x value
						NaN, ... 		% refValue
						false, ...		% logarithmic
						false, ...		% show_points
						PLOT_WIDTH, ...
						PLOT_HEIGHT, ...
						false, ... 	% showDepthIntegrated
						false, ...	% showExtrema
						true, ... 	% showR2P2
						NaN, ...	% data_colors
						strcat(outputDir,'/plot_data/predicted_metabolite_concentrations.tsv'));
				
		fprintf('Plotting SMF-predicted reaction rates..\n');
		rates_profiles = zeros(NZ,NR);
		for h=1:NH
			[~,grid_point] = min(abs(SMF_DEPTHS-hotspots(h)));
			rates_profiles(grid_point,:) = rates_profiles(grid_point,:) + reaction_rates(:,h)';
		end
		plotProfiles(	rates_profiles,...
						SMF_DEPTHS, ...
						{}, ... 				% data
						NaN, ...
						[SMF_DEPTHS(1),SMF_DEPTHS(end)], ...
						reaction_names, ...
						'reaction rate ($\rm kmol/(m^2\cdot day)$)', ...
						strcat(outputDir,'/predicted_reaction_rates.pdf'),...
						sprintf('SMF-predicted area-specific reaction rates (stress=%g)',stress), ...
						NaN,... 			% min x value
						NaN, ... 			% refValue
						false, ...			% logarithmic
						false, ...			% show_points
						PLOT_WIDTH, ...
						PLOT_HEIGHT, ...
						false, ... 	% showDepthIntegrated
						false, ...	% showExtrema
						false, ... 	% showR2P2
						NaN, ...	% data_colors
						strcat(outputDir,'/plot_data/predicted_reaction_rates.tsv'));		
	end
	
	fprintf('Finished. All outputs are in directory "%s"\n',outputDir);
	diary off;
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% AUXILIARY FUNCTIONS


% Get file path to a directory for which there exists no similarly named directory or file
% dir names are chosen incrementally, basename001, basename002 and so on, until no basenameXXX* exists
function filepath = getNonExistingSimilarDir(parent, basename)
	% get all items (level 1) in parent dir
	listing  = dir(sprintf('%s/%s*',parent,basename));
	existing = cell(length(listing),1);
	for l=1:length(listing)
		existing{l} = listing(l).name;
	end
	% increment counter until no similarly named item exists
	n = 1;
	while(true)
		filename = sprintf('%s%03d',basename,n);		
		if(not(any(strncmpi(existing,filename,length(filename)))))
			break;
		end
		n = n+1;
	end
	filepath = sprintf('%s/%s',parent,filename);
end


% create a directory, if it does not exit
function makedir(path)
	if ~exist(path, 'dir')
		mkdir(path);
	end
end



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


